﻿using System;
using System.Text;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Drawing2D;
using System.Reflection;
using System.Globalization;
using System.Windows.Forms;

using plexdata.Controls;
using plexdata.CodePage;

namespace plexdata.CharacterLookup
{
    public partial class MainForm : Form
    {
        private enum FormatTypes
        {
            DEC, // Decimal format.
            HEX, // Hexadecimal format.
            UNI, // Unicode format.
            BIN, // Binary format.
            DEF  // Default/plain format.
        }

        private const int ABOUT_MENU_ID = 0x0815;

        private EncodingsHelper[] standardEncodings = null;

        public MainForm()
            : base()
        {
            this.InitializeComponent();
            this.SetupControls();
            this.SetupCharacterTable();
            this.SetupEncodings();
        }

        private void SetupControls()
        {
            switch (CharacterConverter.Instance.MappingType)
            {
                case CharacterConverter.MappingTypes.Mapped:
                    this.chkCharacterMapping.CheckState = CheckState.Checked;
                    break;
                case CharacterConverter.MappingTypes.Controls:
                    this.chkCharacterMapping.CheckState = CheckState.Indeterminate;
                    break;
                case CharacterConverter.MappingTypes.Unmapped:
                    this.chkCharacterMapping.CheckState = CheckState.Unchecked;
                    break;
            }

            this.InputType = NumberBox.Types.DEF;

            this.chkUseUpperCases.Checked = this.numCharInput.Upper;

            this.numCharInput.Clear();
        }

        private void SetupCharacterTable()
        {
            this.lstCharacterTable.Columns.Clear();

            this.lstCharacterTable.Columns.AddRange(new ColumnHeader[]{
                new ColumnHeader(), // Use a dummy column to have right aligned text!
                new ColumnHeader(), new ColumnHeader(), new ColumnHeader(), 
                new ColumnHeader(), new ColumnHeader()
            });

            this.lstCharacterTable.Columns[0].Width = 0;
            this.lstCharacterTable.Columns[0].TextAlign = HorizontalAlignment.Right; // Don't care...

            this.lstCharacterTable.Columns[1].Width = 60;
            this.lstCharacterTable.Columns[1].TextAlign = HorizontalAlignment.Right;

            this.lstCharacterTable.Columns[2].Width = 80;
            this.lstCharacterTable.Columns[2].TextAlign = HorizontalAlignment.Center;

            this.lstCharacterTable.Columns[3].Width = 80;
            this.lstCharacterTable.Columns[3].TextAlign = HorizontalAlignment.Center;

            this.lstCharacterTable.Columns[4].Width = 150;
            this.lstCharacterTable.Columns[4].TextAlign = HorizontalAlignment.Center;

            this.lstCharacterTable.Columns[5].Width = 80;
            this.lstCharacterTable.Columns[5].TextAlign = HorizontalAlignment.Center;

            this.lstCharacterTable.View = this.chkShowAsTable.Checked ? View.LargeIcon : View.Details;
            this.lstCharacterTable.OwnerDraw = this.chkShowAsTable.Checked;

            this.lstCharacterTable.VirtualMode = true;
            this.lstCharacterTable.VirtualListSize = 0;
            this.lstCharacterTable.RetrieveVirtualItem += new RetrieveVirtualItemEventHandler(CharacterTable_RetrieveVirtualItem);
            this.lstCharacterTable.DrawItem += new DrawListViewItemEventHandler(CharacterTable_DrawItem);
            this.lstCharacterTable.SizeChanged += new EventHandler(CharacterTable_SizeChanged);

            // Fake to adjust item size;
            this.lstCharacterTable.LargeImageList = new ImageList();
            if (Environment.OSVersion.Version.Major >= 6) // Vista and higher.
            {
                this.lstCharacterTable.LargeImageList.ImageSize = new Size(1, 22);
            }
            else
            {
                this.lstCharacterTable.LargeImageList.ImageSize = new Size(40, 22);
            }

            // Needed for flicker free drawing while resizing!
            PropertyInfo property = typeof(ListView).GetProperty("DoubleBuffered", BindingFlags.NonPublic | BindingFlags.Instance);
            property.SetValue(this.lstCharacterTable, true, null);
        }

        private void SetupEncodings()
        {
            this.cmbEncodings.BeginUpdate();
            try
            {
                EncodingsHelper[] entries = null;
                EncodingsHelper selected = this.GetSelectedEncoding();

                // Clear current entries.
                this.cmbEncodings.Items.Clear();

                // Always add default encoding.
                this.cmbEncodings.Items.Add(new EncodingsHelper());

                // Use default selection if nothing is selected yet.
                if (selected == null) { selected = (EncodingsHelper)this.cmbEncodings.Items[0]; }

                if (this.chkUseSystemEncodings.Checked)
                {
                    // Use all available encodings.
                    EncodingInfo[] encodings = Encoding.GetEncodings();
                    entries = new EncodingsHelper[encodings.Length];
                    for (int index = 0; index < encodings.Length; index++)
                    {
                        entries[index] = new EncodingsHelper(encodings[index].GetEncoding());
                    }

                    // Sort by names!
                    this.cmbEncodings.Sorted = true;
                }
                else
                {
                    if (this.standardEncodings == null)
                    {
                        #region Setup Standard Encodings.

                        this.standardEncodings = new EncodingsHelper[]
                        {
                            new EncodingsHelper(
                                new ASCIIDOSLatinUS(), 
                                "Standard ASCII Codepage for MS-DOS (US).",
                                "DOS Latin US"
                                ),

                            new EncodingsHelper(
                                new ASCIIDOSLatin1(), 
                                "Standard ASCII Codepage for MS-DOS (Latin 1).",
                                "DOS Latin 1"
                                ),

                            new EncodingsHelper(
                                new ASCIIDOSLatin2(), 
                                "Standard ASCII Codepage for MS-DOS (Latin 2).",
                                "DOS Latin 2"
                                ),

                            new EncodingsHelper(
                                Encoding.GetEncoding(CultureInfo.CurrentCulture.TextInfo.ANSICodePage), 
                                "American National Standards Institute (ANSI).",
                                "Current ANSI Codepage"
                                ),

                            new EncodingsHelper(
                                Encoding.GetEncoding(CultureInfo.CurrentCulture.TextInfo.OEMCodePage), 
                                "Original Equipment Manufacturer (OEM).",
                                "Current OEM Codepage"
                                ),

                            new EncodingsHelper(
                                new ASCIIEncoding(),
                                "American Standard Code for Information Interchange (ASCII).",
                                "Standard ASCII Codepage"
                                ),

                            new EncodingsHelper(
                                Encoding.GetEncoding(437),
                                "American Standard Code for Information Interchange (ASCII), Codepage=437.",
                                "ASCII Codepage 437"
                                ),

                            new EncodingsHelper(
                                Encoding.GetEncoding(850),
                                "American Standard Code for Information Interchange (ASCII), Codepage=850.",
                                "ASCII Codepage 850"
                                ),

                            new EncodingsHelper(
                                Encoding.GetEncoding(1252),
                                "American National Standards Institute (ANSI), Codepage=1252.",
                                "ANSI Codepage 1252"
                                ),

                                
                            new EncodingsHelper(
                                new UnicodeEncoding(),
                                "Standard Unicode codepage (UTF-16).",
                                "Unicode (UTF-16)"
                                ),

                            new EncodingsHelper(
                                new UTF32Encoding(),
                                "Standard UTF-32 codepage.",
                                "Unicode (UTF-32)"
                                ),


                            new EncodingsHelper(
                                new UTF8Encoding(),
                                "Standard UTF-8 codepage.",
                                "Unicode (UTF-8)"
                                ),


                            new EncodingsHelper(
                                new UTF7Encoding(),
                                "Standard UTF-7 codepage.",
                                "Unicode (UTF-7)"
                                ),
                            };
                        #endregion // Setup Standard Encodings.
                    }

                    // Just use standard encodings.
                    entries = this.standardEncodings;

                    // Don't sort by names!
                    this.cmbEncodings.Sorted = false;
                }

                // Add current encodings to the combo box.
                this.cmbEncodings.Items.AddRange(entries);

                // Pre-select default encoding!
                this.cmbEncodings.SelectedItem = Encoding.Default;

                // Try selecting previously selected item.
                foreach (object current in this.cmbEncodings.Items)
                {
                    if (selected.Equals(current))
                    {
                        this.cmbEncodings.SelectedItem = current;
                        break;
                    }
                }
            }
            catch (Exception exception)
            {
                System.Diagnostics.Debug.WriteLine(exception);
            }
            finally
            {
                this.cmbEncodings.EndUpdate();
            }
        }

        private EncodingsHelper GetSelectedEncoding()
        {
            try
            {
                if (this.cmbEncodings.SelectedItem != null)
                {
                    return (EncodingsHelper)this.cmbEncodings.SelectedItem;
                }
            }
            catch
            {
            }
            return new EncodingsHelper(Encoding.Default);
        }

        private void UpdateGeneralStatus()
        {
            try
            {
                this.stpGeneralLabel.Text = this.GetSelectedEncoding().Description;
            }
            catch
            {
                this.stpGeneralLabel.Text = "";
            }
        }

        private void UpdateCharacterTable()
        {
            this.lstCharacterTable.BeginUpdate();
            try
            {
                int selected = this.GetSelectedListItem();

                if (this.GetSelectedEncoding().Equals(Encoding.ASCII))
                {
                    // Pure ASCII only supports 128 characters!
                    this.lstCharacterTable.VirtualListSize = 128;
                }
                else
                {
                    // Get list view limitation value.
                    int limit = this.GetSelectedEncoding().Encoding.GetByteCount(new char[] { '\u0000' });

                    // Set list view's new virtual size.
                    this.lstCharacterTable.VirtualListSize = (limit == 1) ? 256 : 65536;
                }

                this.SetSelectedListItem(selected);
            }
            catch
            {
            }
            finally
            {
                this.lstCharacterTable.EndUpdate();
            }
        }

        private int GetSelectedListItem()
        {
            if (this.lstCharacterTable.SelectedIndices.Count > 0)
            {
                return this.lstCharacterTable.SelectedIndices[0];
            }
            else
            {
                return -1;
            }
        }

        private void SetSelectedListItem(int selected)
        {
            if (selected != -1)
            {
                if (selected >= this.lstCharacterTable.VirtualListSize)
                {
                    selected = this.lstCharacterTable.VirtualListSize - 1;
                }
                this.lstCharacterTable.Items[selected].EnsureVisible();
                this.lstCharacterTable.Items[selected].Selected = true;
            }
        }

        private void DelSelectedListItem()
        {
            foreach (int index in this.lstCharacterTable.SelectedIndices)
            {
                this.lstCharacterTable.Items[index].Selected = false;
            }
        }

        private string GetCellText(int value, FormatTypes format)
        {
            try
            {
                switch (format)
                {
                    case FormatTypes.DEC:

                        return String.Format("{0}", System.Convert.ToString(value, 10));

                    case FormatTypes.HEX:

                        return String.Format("0x{0}", System.Convert.ToString(value, 16).PadLeft(4, '0').ToUpper());

                    case FormatTypes.UNI:

                        Encoding encoding = GetSelectedEncoding().Encoding;
                        if (encoding is ASCIICodePageBase)
                        {
                            value = ((ASCIICodePageBase)encoding).CharacterFromByte((byte)value);
                        }
                        return String.Format("\\u{0}", System.Convert.ToString(value, 16).PadLeft(4, '0').ToUpper());

                    case FormatTypes.BIN:

                        if (this.lstCharacterTable.VirtualListSize <= 256)
                        {
                            return String.Format("{0}", System.Convert.ToString(value, 2).PadLeft(8, '0'));
                        }
                        else
                        {
                            return String.Format("{0}", System.Convert.ToString(value, 2).PadLeft(16, '0'));
                        }

                    case FormatTypes.DEF:

                        return CharacterConverter.Instance.Encode(value, this.GetSelectedEncoding().Encoding);
                }
            }
            catch
            {
            }
            return "<error>";
        }

        private string GetItemToolTip(int item)
        {
            return String.Format(
                "DEC: {0}\nHEX: {1}\nUNI: {2}\nBIN: {3}",
                this.GetCellText(item, FormatTypes.DEC),
                this.GetCellText(item, FormatTypes.HEX),
                this.GetCellText(item, FormatTypes.UNI),
                this.GetCellText(item, FormatTypes.BIN)
            );
        }

        private NumberBox.Types InputType
        {
            get
            {
                if (this.btnInputType.Tag != null)
                {
                    return (NumberBox.Types)this.btnInputType.Tag;
                }
                else
                {
                    return NumberBox.Types.DEF;
                }
            }
            set
            {
                switch (value)
                {
                    case NumberBox.Types.DEC:
                    case NumberBox.Types.HEX:
                    case NumberBox.Types.BIN:
                        break;
                    default:
                        value = NumberBox.Types.DEF;
                        break;
                }
                this.btnInputType.Text = value.ToString();
                this.btnInputType.Tag = value;
            }
        }

        public new Point Location
        {
            get { return base.Location; }
            set
            {
                Rectangle helper = new Rectangle(value, base.Size);
                foreach (Screen screen in Screen.AllScreens)
                {
                    if (screen.Bounds.Contains(helper))
                    {
                        base.Location = value;
                        return;
                    }
                }
            }
        }

        public new Size Size
        {
            get { return base.Size; }
            set
            {
                if (value.Height >= base.MinimumSize.Height && value.Width >= base.MinimumSize.Width)
                {
                    base.Size = value;
                }
            }
        }

        protected override void WndProc(ref Message message)
        {
            // Check for incoming system commands...
            if (message.Msg == SystemMenu.WM_SYSCOMMAND)
            {
                // Check if About box menu item was clicked...
                switch (message.WParam.ToInt32())
                {
                    case ABOUT_MENU_ID:
                        AboutBox aboutBox = new AboutBox();
                        aboutBox.ShowDialog(this);
                        return;
                }
            }

            base.WndProc(ref message);
        }

        #region Window and Control event handler.

        private void MainForm_Load(object sender, EventArgs args)
        {
            ApplicationSettings settings;

            if (!plexdata.Generic.Serializer.Load<ApplicationSettings>(Program.SettingsFilename, out settings))
            {
                settings = new ApplicationSettings(); // Use default values;
            }

            if (settings.InputType != null)
            {
                this.InputType = (NumberBox.Types)settings.InputType;
            }

            this.numCharInput.Type = this.InputType;
            this.chkUseUpperCases.CheckState = settings.UseUpperCases;
            this.chkUseSystemEncodings.CheckState = settings.UseSystemEncodings;
            this.chkCharacterMapping.CheckState = settings.CharacterMapping;
            this.chkShowAsTable.CheckState = settings.ShowAsTable;
            this.cmbEncodings.SelectedItem = settings.Encoding;
            this.Size = settings.Size;
            this.Location = settings.Location;

            // Append About box menu item to Form's system menu.
            SystemMenu menu = SystemMenu.FromWindow(this);
            menu.AddSeparator();
            menu.AddItem(ABOUT_MENU_ID, "About...");
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs args)
        {
            ApplicationSettings settings = new ApplicationSettings();

            settings.InputType = this.InputType;
            settings.UseUpperCases = this.chkUseUpperCases.CheckState;
            settings.UseSystemEncodings = this.chkUseSystemEncodings.CheckState;
            settings.CharacterMapping = this.chkCharacterMapping.CheckState;
            settings.ShowAsTable = this.chkShowAsTable.CheckState;
            settings.Encoding = this.GetSelectedEncoding().Encoding;
            settings.Size = this.Size;
            settings.Location = this.Location;

            plexdata.Generic.Serializer.Save<ApplicationSettings>(Program.SettingsFilename, settings);
        }

        private void Encodings_SelectedIndexChanged(object sender, EventArgs args)
        {
            this.UpdateGeneralStatus();
            this.UpdateCharacterTable();
        }

        private void UseSystemEncodings_CheckedChanged(object sender, EventArgs args)
        {
            int selected = GetSelectedListItem();
            this.SetupEncodings();
            this.SetSelectedListItem(selected);
        }

        private void Encodings_MouseHover(object sender, EventArgs args)
        {
            try
            {
                this.tipGeneral.SetToolTip(this.cmbEncodings, this.GetSelectedEncoding().Tooltip);
            }
            catch
            {
            }
        }

        private void CharacterTable_RetrieveVirtualItem(object sender, RetrieveVirtualItemEventArgs args)
        {
            try
            {
                int index = args.ItemIndex;

                args.Item = new ListViewItem(new string[]{
                    null,
                    this.GetCellText(index, FormatTypes.DEC),
                    this.GetCellText(index, FormatTypes.HEX),
                    this.GetCellText(index, FormatTypes.UNI),
                    this.GetCellText(index, FormatTypes.BIN),
                    this.GetCellText(index, FormatTypes.DEF)
                });
                args.Item.ToolTipText = this.GetItemToolTip(index);
            }
            catch
            {
                args.Item = null;
            }
        }

        private void CharacterTable_SizeChanged(object sender, EventArgs args)
        {
            this.lstCharacterTable.Invalidate(true);
        }

        private void CharacterTable_DrawItem(object sender, DrawListViewItemEventArgs args)
        {
            Bitmap bitmap = null;
            Graphics offscreen = null;
            Rectangle bounds = default(Rectangle);
            Color foreColor = args.Item.ForeColor; // Do not change this value inside this function!
            Color backColor = args.Item.BackColor; // Do not change this value inside this function!

            try
            {
                // Create a bitmap the size of the form.
                bitmap = new Bitmap(args.Bounds.Width, args.Bounds.Height);

                // Create a Graphics object that is not on the form.
                offscreen = Graphics.FromImage(bitmap);

                // Create offscreen bounding rectangle.
                bounds = new Rectangle(0, 0, bitmap.Width, bitmap.Height);

                if (args.Item.Selected)
                {
                    if (this.lstCharacterTable.Focused)
                    {
                        // Fill bounding rectangle with "selected" background color.
                        offscreen.FillRectangle(SystemBrushes.MenuHighlight, bounds);

                        // Don't forget to draw focus rectangle!
                        ControlPaint.DrawFocusRectangle(offscreen, bounds, backColor, foreColor);

                        // Adjust item's foreground color!
                        foreColor = SystemColors.HighlightText;
                    }
                    else
                    {
                        if (this.lstCharacterTable.HideSelection)
                        {
                            // Fill bounding rectangle with "normal" background color.
                            offscreen.FillRectangle(new SolidBrush(backColor), bounds);
                        }
                        else
                        {
                            // Fill bounding rectangle with "un-focused" background color.
                            offscreen.FillRectangle(SystemBrushes.Control, bounds);
                        }
                    }
                }
                else
                {
                    // Fill bounding rectangle with "normal" background color.
                    offscreen.FillRectangle(new SolidBrush(backColor), bounds);
                }

                // Prepare bounding rectangle for text output.
                bounds.Inflate(-2, -2);

                // Adjust text rendering behaviour.
                SmoothingMode oldMode = offscreen.SmoothingMode;
                offscreen.SmoothingMode = SmoothingMode.AntiAlias;

                TextRenderingHint oldHint = offscreen.TextRenderingHint;
                offscreen.TextRenderingHint = TextRenderingHint.AntiAlias;

                // Adjust text render flags.
                TextFormatFlags flags = TextFormatFlags.Top | TextFormatFlags.HorizontalCenter | TextFormatFlags.LeftAndRightPadding;

                // Draw current character.
                TextRenderer.DrawText(
                    offscreen,
                    this.GetCellText(args.ItemIndex, FormatTypes.DEF),
                    new Font(args.Item.Font.FontFamily, 14F),
                    bounds,
                    foreColor,
                    flags
                    );

                // Restore previous text rendering behaviour.
                offscreen.SmoothingMode = oldMode;
                offscreen.TextRenderingHint = oldHint;

                // Adjust output string format to bottom/center.
                StringFormat format = new StringFormat(StringFormatFlags.NoWrap);
                format.Alignment = StringAlignment.Center;
                format.LineAlignment = StringAlignment.Far;

                // Draw current index.
                offscreen.DrawString(
                    this.GetCellText(args.ItemIndex, FormatTypes.DEC),
                    new Font(args.Item.Font.FontFamily, 8F),
                    new SolidBrush(foreColor),
                    bounds,
                    format
                    );

                // Perform bit-block transfer onto destination graphics context.
                args.Graphics.DrawImage(bitmap, args.Bounds);
            }
            finally
            {
                // Cleanup!
                if (offscreen != null) { offscreen.Dispose(); }
                if (bitmap != null) { bitmap.Dispose(); }
            }
        }

        private void CharacterMapping_Click(object sender, EventArgs args)
        {
            switch (this.chkCharacterMapping.CheckState)
            {
                case CheckState.Checked:
                    this.chkCharacterMapping.CheckState = CheckState.Indeterminate;
                    break;
                case CheckState.Indeterminate:
                    this.chkCharacterMapping.CheckState = CheckState.Unchecked;
                    break;
                case CheckState.Unchecked:
                    this.chkCharacterMapping.CheckState = CheckState.Checked;
                    break;
            }
        }

        private void CharacterMapping_CheckStateChanged(object sender, EventArgs args)
        {
            switch (this.chkCharacterMapping.CheckState)
            {
                case CheckState.Checked:
                    CharacterConverter.Instance.MappingType = CharacterConverter.MappingTypes.Mapped;
                    break;
                case CheckState.Indeterminate:
                    CharacterConverter.Instance.MappingType = CharacterConverter.MappingTypes.Controls;
                    break;
                case CheckState.Unchecked:
                    CharacterConverter.Instance.MappingType = CharacterConverter.MappingTypes.Unmapped;
                    break;
            }

            this.lstCharacterTable.Invalidate(true);
        }

        private void CharInput_TypeChanged(object sender, EventArgs args)
        {
            this.InputType = this.numCharInput.Type;
        }

        private void CharInput_ValueChanged(object sender, EventArgs args)
        {
            try
            {
                int selected = Convert.ToInt32(this.numCharInput.Value);
                if (selected < this.lstCharacterTable.VirtualListSize)
                {
                    this.SetSelectedListItem(selected);
                }
                else
                {
                    this.DelSelectedListItem();
                }
            }
            catch
            {
                this.DelSelectedListItem();
            }
        }

        private void InputType_Click(object sender, EventArgs args)
        {
            switch (this.InputType)
            {
                case NumberBox.Types.DEC:
                    this.numCharInput.Type = NumberBox.Types.HEX;
                    break;
                case NumberBox.Types.HEX:
                    this.numCharInput.Type = NumberBox.Types.BIN;
                    break;
                case NumberBox.Types.BIN:
                    this.numCharInput.Type = NumberBox.Types.DEC;
                    break;
                default:
                    throw new IndexOutOfRangeException("Unexpected type!");
            }
        }

        private void UseUpperCases_CheckedChanged(object sender, EventArgs args)
        {
            this.numCharInput.Upper = this.chkUseUpperCases.Checked;
        }

        private void ShowAsTable_CheckedChanged(object sender, EventArgs args)
        {
            this.lstCharacterTable.View = this.chkShowAsTable.Checked ? View.LargeIcon : View.Details;
            this.lstCharacterTable.OwnerDraw = this.chkShowAsTable.Checked;
            this.UpdateCharacterTable();
        }

        private void CharacterTable_SelectedIndexChanged(object sender, EventArgs args)
        {
            if (this.lstCharacterTable.SelectedIndices.Count > 0)
            {
                this.numCharInput.Value = this.lstCharacterTable.SelectedIndices[0];
            }
        }

        #endregion // Window and Control event handler.
    }

    internal class EncodingsHelper
    {
        private string display = "";
        private string tooltip = "";
        private Encoding encoding = null;

        internal EncodingsHelper()
            : this(null)
        {
            this.tooltip = "\"" + this.Display + "\" is used as default encoding.";
            this.display = "(Default)";
        }

        internal EncodingsHelper(Encoding encoding)
            : base()
        {
            this.encoding = (encoding == null) ? Encoding.Default : encoding;
            this.display = this.Display;
            this.tooltip = this.Tooltip;
        }

        internal EncodingsHelper(Encoding encoding, string tooltip)
            : this(encoding)
        {
            this.tooltip = tooltip;
        }

        internal EncodingsHelper(Encoding encoding, string tooltip, string display)
            : this(encoding, tooltip)
        {
            this.display = display;
        }

        internal EncodingsHelper(int codepage)
            : this(Encoding.GetEncoding(codepage))
        {
        }

        internal EncodingsHelper(int codepage, string tooltip)
            : this(codepage)
        {
            this.tooltip = tooltip;
        }

        internal EncodingsHelper(int codepage, string tooltip, string display)
            : this(codepage, tooltip)
        {
            this.display = display;
        }

        internal Encoding Encoding
        {
            get { return this.encoding; }
        }

        internal string Display
        {
            get
            {
                if (String.IsNullOrEmpty(this.display))
                {
                    return this.encoding.EncodingName;
                }
                else
                {
                    return this.display;
                }
            }
        }

        internal string Tooltip
        {
            get
            {
                if (String.IsNullOrEmpty(this.tooltip))
                {
                    return this.Description;
                }
                else
                {
                    return this.tooltip;
                }
            }
        }

        internal string Description
        {
            get
            {
                if (this.encoding.CodePage == this.encoding.WindowsCodePage)
                {
                    return String.Format("{0} [{1}]", this.encoding.EncodingName, this.encoding.CodePage);
                }
                else
                {
                    return String.Format("{0} [{1}, {2}]", this.encoding.EncodingName, this.encoding.CodePage, this.encoding.WindowsCodePage);
                }
            }
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        public override bool Equals(object other)
        {
            if (other is int)
            {
                return (this.encoding.CodePage == (int)other);
            }
            else if (other is Encoding)
            {
                return (this.encoding.GetHashCode() == ((Encoding)other).GetHashCode());
            }
            else if (other is EncodingsHelper)
            {
                return (this.encoding.GetHashCode() == ((EncodingsHelper)other).encoding.GetHashCode());
            }
            else
            {
                return base.Equals(other);
            }
        }

        public override string ToString()
        {
            if (!String.IsNullOrEmpty(this.display))
            {
                return this.display;
            }
            else if (this.encoding != null)
            {
                return this.encoding.EncodingName;
            }
            else
            {
                return base.ToString();
            }
        }
    }
}
