﻿using System;
using System.Windows.Forms;
using System.ComponentModel;

namespace plexdata.Controls
{
    public class NumberBox : TextBox
    {
        public enum Types { DEC, HEX, BIN, DEF=DEC }

        public event EventHandler<EventArgs> TypeChanged;
        public event EventHandler<EventArgs> ValueChanged;

        public delegate void TypeChangedEventHandler(object sender, EventArgs args);
        public delegate void ValueChangedEventHandler(object sender, EventArgs args);

        private Types type;
        private bool upper;

        public NumberBox()
            : base()
        {
            this.TextAlign = HorizontalAlignment.Right;
            this.Value = 0;
            this.Type = Types.DEF;
            this.Upper = true;

            base.TextChanged += new EventHandler(NumberBox_TextChanged);
        }

        [Browsable(true)]
        [DefaultValue(Types.DEF)]
        [Description("Indicates the type how the input data should be displayed.")]
        public Types Type
        {
            get { return this.type; }
            set
            {
                string current = "";
                if (this.type != value)
                {
                    if (base.TextLength > 0)
                    {
                        // Convert from source type...
                        long helper = this.AsValue(base.Text, this.type, base.Text[0] == '-');

                        // ...into destination type.
                        current = this.AsString(helper, value, this.Upper);
                    }

                    // Change current type anyway!
                    this.type = value;

                    // Signal type changes to all listeners.
                    this.RaiseTypeChanged();

                    // Now change inherited text field!
                    if (!String.IsNullOrEmpty(current))
                    {
                        base.Text = current;
                    }
                }
            }
        }

        [Browsable(true)]
        [DefaultValue(true)]
        [Description("If true than display HEX values in upper cases.")]
        public bool Upper
        {
            get { return this.upper; }
            set
            {
                if (this.upper != value)
                {
                    if (value)
                    {
                        base.Text = base.Text.ToUpper();
                    }
                    else
                    {
                        base.Text = base.Text.ToLower();
                    }

                    // Change current upper value!
                    this.upper = value;
                }
            }
        }

        [Browsable(true)]
        [DefaultValue(0)]
        [Description("The current value of the number text box control.")]
        public long Value
        {
            get
            {
                try
                {
                    return this.AsValue(base.Text, this.Type, base.Text[0] == '-');
                }
                catch (Exception exception)
                {
                    base.Text = "";  // Reset inherited text field and ...
                    throw exception; // ... inform caller about failure!
                }
            }
            set
            {
                try
                {
                    base.Text = this.AsString(value, this.Type, this.Upper);

                    // Signal value changes to all listeners.
                    this.RaiseValueChanged();
                }
                catch
                {
                }
            }
        }

        [DefaultValue(HorizontalAlignment.Right)]
        public new HorizontalAlignment TextAlign
        {
            get { return base.TextAlign; }
            set { base.TextAlign = value; }
        }

        [Browsable(false)] // Do not show in the designer's property pane!
        public new string Text
        {
            get { return ""; }
            set { }
        }

        [Browsable(false)] // Do not show in the designer's property pane!
        public override bool Multiline
        {
            get { return false; }
            set { }
        }

        [Browsable(false)] // Do not show in the designer's property pane!
        public new string[] Lines
        {
            get { return new string[] { base.Text }; }
            set { }
        }

        private string AsString(long value, Types type, bool toUpper)
        {
            string result = "";
            switch (type)
            {
                case Types.DEC:
                    result = System.Convert.ToString(value, 10);
                    break;
                case Types.HEX:
                    result = System.Convert.ToString(value, 16);
                    break;
                case Types.BIN:
                    result = System.Convert.ToString(value, 2);
                    break;
                default:
                    throw new NotSupportedException("Given number type \"" + type + "\" is not supported!");
            }

            if (toUpper)
            {
                return result.ToUpper();
            }
            else
            {
                return result.ToLower();
            }
        }

        private long AsValue(string text, Types type)
        {
            switch (type)
            {
                case Types.DEC:
                    return System.Convert.ToInt64(text, 10);
                case Types.HEX:
                    return System.Convert.ToInt64(text, 16);
                case Types.BIN:
                    return System.Convert.ToInt64(text, 2);
                default:
                    throw new NotSupportedException("Given number type \"" + type + "\" is not supported!");
            }
        }

        private long AsValue(string text, Types type, bool negative)
        {
            try
            {
                return this.AsValue(text, type);
            }
            catch (OverflowException)
            {
                return (negative && this.Type == Types.DEC) ? Int64.MinValue : Int64.MaxValue;
            }
        }

        private bool IsValidInput(KeyEventArgs args)
        {
            if (this.Type == Types.DEC)
            {
                return this.IsDecInput(args);
            }
            else if (this.Type == Types.HEX)
            {
                return this.IsHexInput(args);
            }
            else if (this.Type == Types.BIN)
            {
                return this.IsBinInput(args);
            }
            else
            {
                return false;
            }
        }

        private bool IsDecInput(KeyEventArgs args)
        {
            // Check for number digit.
            if (('0' <= args.KeyValue && args.KeyValue <= '9') ||
                (Keys.NumPad0 <= args.KeyCode && args.KeyCode <= Keys.NumPad9))
            {
                // Don't allow modifiers!
                return !(args.Shift || args.Alt || args.Control);
            }
            // Check for number sign!
            else if (args.KeyCode == Keys.Subtract || args.KeyCode == Keys.OemMinus)
            {
                if (base.TextLength == 0 || (base.SelectionStart == 0 && !base.Text.Contains("-")))
                {
                    // Don't allow modifiers!
                    return !(args.Shift || args.Alt || args.Control);
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return this.IsControlKey(args);
            }
        }

        private bool IsHexInput(KeyEventArgs args)
        {
            // Check for number digit.
            if (('0' <= args.KeyValue && args.KeyValue <= '9') ||
                (Keys.NumPad0 <= args.KeyCode && args.KeyCode <= Keys.NumPad9))
            {
                // Don't allow modifiers!
                return !(args.Shift || args.Alt || args.Control);
            }
            // Check for HEX letter digit.
            else if ('A' <= args.KeyValue && args.KeyValue <= 'F')
            {
                // Allow SHIFT modifier only because of upper cases!
                return !(args.Alt || args.Control);
            }
            else
            {
                return this.IsControlKey(args);
            }
        }

        private bool IsBinInput(KeyEventArgs args)
        {
            // Check for binary digit.
            if (('0' == args.KeyValue || args.KeyValue == '1') ||
                (Keys.NumPad0 == args.KeyCode || args.KeyCode == Keys.NumPad1))
            {
                // Don't allow modifiers!
                return !(args.Shift || args.Alt || args.Control);
            }
            else
            {
                return this.IsControlKey(args);
            }
        }

        private bool IsControlKey(KeyEventArgs args)
        {
            if (args.KeyValue < '0')
            {
                // Allow each control key except SPACE!
                return (args.KeyCode != Keys.Space);
            }
            else
            {
                // Check for Copy & Paste keys (CTRL-C, CTRL-X and CTRL-V)
                if (args.Control && !args.Alt && !args.Shift)
                {
                    if (args.KeyValue == 'C' || args.KeyValue == 'X' || args.KeyValue == 'V')
                    {
                        base.SelectAll(); // Select complete text box content!
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                else
                {
                    return true;
                }
            }
        }

        protected override void OnKeyDown(KeyEventArgs args)
        {
            args.SuppressKeyPress = !this.IsValidInput(args);
            base.OnKeyDown(args);
        }

        protected override void OnKeyPress(KeyPressEventArgs args)
        {
            if (this.Type == Types.HEX && this.Upper)
            {
                if ('a' <= args.KeyChar && args.KeyChar <= 'f')
                {
                    args.KeyChar = (char)('A' + (args.KeyChar - 'a'));
                }
            }
            base.OnKeyPress(args);
        }

        private void NumberBox_TextChanged(object sender, EventArgs args)
        {
            // Signal value changes to all listeners.
            this.RaiseValueChanged();
        }

        protected virtual void RaiseTypeChanged()
        {
            EventHandler<EventArgs> handler = TypeChanged;
            if (handler != null)
            {
                handler(this, new EventArgs());
            }
        }

        protected virtual void RaiseValueChanged()
        {
            EventHandler<EventArgs> handler = ValueChanged;
            if (handler != null)
            {
                handler(this, new EventArgs());
            }
        }

        private void TryPaste(string paste)
        {
            string original = base.Text;
            Types type = this.Type;
            bool upper = this.Upper;
            try
            {
                // First of all try current conversion!
                base.Text = this.AsString(this.AsValue(paste, type), type, upper);
            }
            catch
            {
                try
                {
                    // Try decimal conversion.
                    type = Types.DEC;
                    base.Text = this.AsString(this.AsValue(paste, type), type, upper);
                    if (this.type != type)
                    {
                        this.type = type;
                        this.RaiseTypeChanged();
                    }
                }
                catch
                {
                    try
                    {
                        // Try hexadecimal conversion.
                        type = Types.HEX;
                        base.Text = this.AsString(this.AsValue(paste, type), type, upper);
                        if (this.type != type)
                        {
                            this.type = type;
                            this.RaiseTypeChanged();
                        }
                    }
                    catch
                    {
                        try
                        {
                            // Try binary conversion.
                            type = Types.BIN;
                            base.Text = this.AsString(this.AsValue(paste, type), type, upper);
                            if (this.type != type)
                            {
                                this.type = type;
                                this.RaiseTypeChanged();
                            }
                        }
                        catch
                        {
                            base.Text = original;
                        }
                    }
                }
            }
            base.SelectAll(); // Select complete text box content!
        }

        private const int WM_CUT = 0x0300;
        private const int WM_COPY = 0x0301;
        private const int WM_PASTE = 0x0302;
        private const int WM_CONTEXTMENU = 0x007B;

        protected override void WndProc(ref Message message)
        {
            if (message.Msg == WM_PASTE && Clipboard.ContainsText())
            {
                this.TryPaste(Clipboard.GetText().Trim());
                return; // Avoid original behavior!
            }
            else if (message.Msg == WM_CUT)
            {
                base.SelectAll(); // Select complete text box content!
            }
            else if (message.Msg == WM_COPY)
            {
                base.SelectAll(); // Select complete text box content!
            }
            else if (message.Msg == WM_CONTEXTMENU)
            {
                base.SelectAll(); // Select complete text box content!
            }
            base.WndProc(ref message);
        }
    }
}
